#include "configuration.h"
#include "piece.h"
#include "inputspec.h"
#include "queue.h"

#include <iostream>

using namespace std;


InputSpec   *Configuration::spec;
int         Configuration::maxDistance;



// Make a new, empty configuration.

Configuration::Configuration()
{
    pieces = NULL;
    usesGoal = false;
    
    distance = 0;
    predecessor = NULL;
    movedPiece = -1;
    movedShifts = NULL;
}


// Make a configuration which is a copy of another one.

Configuration::Configuration(Configuration &config)
{
    *this = config;
    
    // Copy the pieces.
    
    if (pieces)
        pieces = new PieceList(pieces);
        
    // Copy the direction list.
    
    if (movedShifts)
        movedShifts = new DirectionList(movedShifts);
}


// Dispose the configuration.

Configuration::~Configuration()
{
    if (pieces)
    {
        pieces->deleteContents();
        delete pieces;
    }
    
    if (movedShifts)
    {
        movedShifts->deleteContents();
        delete movedShifts;
    }
}


// Try all possible piece placements at (x, y) (including 'filling' it with
// empty space), and for each one, recursively try to fill the next spot in
// the box.  If we've thus filled the entire box, we have a candidate puzzle
// layout; see if it can be solved.

void Configuration::puzzleSearch(int x, int y)
{
    if (y < spec->boxHeight)
    {
        // Try all piece placements at (x, y).  For i == NumShapes, we choose to
        // make this mean fill square (x, y) with nothing.
        
        int nextx = (x + 1) % spec->boxWidth;
        int nexty = y + !nextx;
        
        for (int i = 0; i <= spec->numShapes; ++i)
            if (i == spec->numShapes || placePiece(i, x, y))
            {
                puzzleSearch(nextx, nexty);
                
                // Now take the last piece out, so we can try the next.
                
                if (i < spec->numShapes)
                    removePiece();
            }
    }
    else
        solve();                // We've filled the whole box.  Try to solve this puzzle.
}


// Can this configuration be solved by making some sequence of moves?
// If so, and the minimum-length solution is longer than the longest found so
// far, print it out.

void Configuration::solve()
{
    if (!usesGoal)              // Can't solve it if it doesn't even contain the goal piece!
        return;
    
    Queue queue;
    ConfigurationList *seen = NULL;
    
    // Do a BFS from (a copy of) this configuration, looking for a solved state.
    // See CLRS, p. 532.  We don't use an explicit 'color' field; instead, we keep
    // a list 'seen' of all configurations we've already visited.
    
    queue.enqueue(new Configuration(*this));
        
    while (!queue.empty())
    {
        Configuration *config = queue.dequeue();

        seen = new ConfigurationList(config, seen);         // Add this config to the seen list.
        
        // Find all configurations reachable from config by moving a single piece.
        
        PieceList *pl;
        int pieceNum;
        
        ConfigurationList *oldmoves;
        
        for (pl = config->pieces, pieceNum = 0; pl; pl = pl->next, ++pieceNum)
            for (ConfigurationList *moves = config->findMoves(pieceNum); moves; oldmoves = moves, moves = moves->next, delete oldmoves)
            {
                Configuration *newconfig = moves->config;
                
                // Have we already looked at this configuration?
                
                if ((!seen || !seen->isPresent(newconfig)) && !queue.isPresent(newconfig))
                {
                    // Found a new reachable configuration.  Check to see if it's solved, and add it to the queue.
                    
                    newconfig->distance = config->distance + 1;
                    newconfig->predecessor = config;
                    newconfig->movedPiece = pieceNum;
                    
                    if (newconfig->solved())
                    {                           
                        if (newconfig->distance > maxDistance)
                        {
                            // This is the most difficult puzzle we've found yet.
                            
                            maxDistance = newconfig->distance;

                            newconfig->printSolution();
                        }
                        
                        // We know the puzzle can be solved at this depth; no use looking any farther.
                        
                        delete newconfig;
                        
                        goto done;
                    }
                    
                    // The new configuration is not solved; add it to the search queue.
                    
                    queue.enqueue(newconfig);
                }
                else
                    delete newconfig;               
            }
    }
    
done:

    // clean up
    
    if (seen)
    {
        seen->deleteContents();
        delete seen;
    }
    
    while (!queue.empty())
        delete queue.dequeue();
}


// Return a list of all configurations reachable from this one by a sequence of shifts of a given piece.
// To do this, we do another BFS.

ConfigurationList *Configuration::findMoves(int pieceNum)
{
    Queue queue;
    ConfigurationList *seen = NULL;
    
    Configuration *start = new Configuration(*this);
    
    if (start->movedShifts)
    {
        start->movedShifts->deleteContents();
        delete start->movedShifts;
        start->movedShifts = NULL;
    }
    
    queue.enqueue(start);

    while (!queue.empty())
    {
        Configuration *config = queue.dequeue();
        seen = new ConfigurationList(config, seen);
    
        Piece *piece = config->pieces->index(pieceNum);
        
        for (int dir = 0; dir < 4; ++dir)
            if (piece->canSlide(dir, config))
            {
                // We found a piece that can slide.
                
                Configuration *newconfig = new Configuration(*config);
                
                newconfig->pieces->index(pieceNum)->shift(dir);
                
                // Have we not seen this one yet?
                
                if ((!seen || !seen->isPresent(newconfig)) && !queue.isPresent(newconfig))
                {
                    // Found a new reachable configuration.
                    // Add it to the queue.

                    newconfig->movedShifts = new DirectionList(dir, newconfig->movedShifts);
                    queue.enqueue(newconfig);   
                }
                else
                    delete newconfig;           
            }   
    }
    
    return seen;
}


// Is this configuration solved?  It's solved if the goal piece is in the goal
// location.

bool Configuration::solved()
{
    for (PieceList *pl = pieces; pl; pl = pl->next)
        if (pl->piece->solved())
            return true;
            
    return false;
}


// Print the solution.  We recursively call printSolution2 on all predecessor positions,
// until there are no more.  Then we know we're at the starting configuration,
// and can print it.  On the way back up the recursion stack, print out the
// move sequence, which we stored in the configurations.

void Configuration::printSolution()
{
    printSolution2();

    cout << "\nEND SOLUTION" << endl;

}

void Configuration::printSolution2()
{
    if (!predecessor)
    {
        // This is our starting configuration.  Print some useful debugging & visualization stuff,
        // then print the formal solution.
        
        cout << "\n\nFound a solution!  Length is " << maxDistance << ".  Here's what it looks like:" << endl;

        prettyPrint();

        cout << "\n\nBEGIN SOLUTION" << endl;

        pieces->print();
        
        cout << "-\n" << maxDistance << "\n-\n";
    }
    else
    {
        predecessor->printSolution2();
        
        cout << "Move piece " << movedPiece << " ";
        movedShifts->print();
        cout << endl;
    }
}


// Print a configuration for human consumption, in a nice ASCII grid layout.

void Configuration::prettyPrint()
{
    char table[2 * MAXBOXDIM + 3][MAXBOXDIM + 2];

    for (int x = 0; x < 2 * MAXBOXDIM + 3; ++x)
        for (int y = 0; y < MAXBOXDIM + 2; ++y)
            table[x][y] = '*';
            
    PieceList *pl;
    int pieceNum;
    
    for (int x = 0; x < spec->boxWidth; ++x)
        for (int y = 0; y < spec->boxHeight; ++y)
        {
            if (!spec->boxShape[x][y])
            {
                table[2 * x + 1][y + 1] = table[2 * x + 3][y + 1] = ' ';
                table[2 * x + 2][y + 1] = '.';
            }
            
            for (pl = pieces, pieceNum = 0; pl; pl = pl->next, ++pieceNum)
                if (pl->piece->occupied(x, y))
                {
                    table[2 * x + 2][y + 1] = '0' + pieceNum;
                    break;
                }
        }
            
    for (int y = 0; y < spec->boxHeight + 2; ++y)
    {
        for (int x = 0; x < 2 * spec->boxWidth + 3; ++x)
            cout << table[x][y];
            
        cout << endl;
    }
}


// Can a piece of a given shape be placed into this configuration with upper
// left corner at (x, y)?
// If so, place it there & return true.  Otherwise return false.

bool Configuration::placePiece(int shapeIndex, int x, int y)
{   
    // we can only use the goal piece once.
    
    if (shapeIndex == spec->goalIndex && usesGoal)
        return false;

    // ... and don't bother starting it in the goal position!
    
    if (shapeIndex == spec->goalIndex && x == spec->goalX && y == spec->goalY)
        return false;
        
    PieceShape *shape = &(spec->shapes[shapeIndex]);

    for (int i = 0; i < MAXPIECEDIM; ++i)
        for (int j = 0; j < MAXPIECEDIM; ++j)
            if ((*shape)[i][j] && occupied(x + i, y + j))
                return false;

    pieces = new PieceList(new Piece(shapeIndex, x, y), pieces);
    
    if (shapeIndex == spec->goalIndex)
        usesGoal = true;
        
    return true;
}


// Delete the last piece added to this configuration.

void Configuration::removePiece()
{
    if (pieces->piece->isGoal())
        usesGoal = false;
        
    PieceList *oldpieces = pieces;
    pieces = pieces->next;
    
    delete oldpieces;
}


// Is the square (x, y) occupied by any piece in this configuration, or by a
// box wall?

bool Configuration::occupied(int x, int y)
{
    if (x < 0 || y < 0 || x >= MAXBOXDIM || y >= MAXBOXDIM || spec->boxShape[x][y])
        return true;
        
    for (PieceList *pl = pieces; pl; pl = pl->next)
        if (pl->piece->occupied(x, y))
            return true;
    
    return false;
}


// Does this configuration have the same piece shapes and locations as another
// configuration?

bool Configuration::sameAs(Configuration *config)
{
    for (PieceList *pl1 = pieces, *pl2 = config->pieces; pl1 || pl2; pl1 = pl1->next, pl2 = pl2->next)
    {
        if (!pl1 || !pl2)
            return false;                       // one list was shorter than the other
            
        if (!pl1->piece->sameAs(pl2->piece))
            return false;                       // corresponding pieces don't match
    }
    
    return true;
}


// Make a new ConfigurationList node.

ConfigurationList::ConfigurationList(Configuration *c, ConfigurationList *cl)
{
    config = c;
    next = cl;
}


// Dispose the contents of the list.

void ConfigurationList::deleteContents()
{
    delete config;
    
    if (next)
    {
        next->deleteContents();
        delete next;
    }
}


// Does any configuration in this list match a given configuration?

bool ConfigurationList::isPresent(Configuration *match)
{
    for (ConfigurationList *cl = this; cl; cl = cl->next)
        if (cl->config->sameAs(match))
            return true;
            
    return false;
}


// Make a new direction list node.
// Direction lists are used for storing the shift sequences that individual pieces do
// to make up a move.

DirectionList::DirectionList(int d, DirectionList *dl)
{
    dir = d;
    next = dl;
}


// Make a copy of an existing direction list.

DirectionList::DirectionList(DirectionList *dl)
{
    dir = dl->dir;
    next = dl->next ? new DirectionList(dl->next) : NULL;
}


// Dispose the contents of the list.

void DirectionList::deleteContents()
{
    if (next)
    {
        next->deleteContents();
        delete next;
    }
}


// Print out the sequence of directions.  We do it in reverse order, since the
// most recent shift is at the head of the list.

void DirectionList::print()
{
    static char *dirNames[] = { "right", "down", "left", "up" };

    if (next)
        next->print();
    
    cout << dirNames[dir] << " ";
}
